import json
import logging
from openerp import _
from openerp import api
from openerp import exceptions
from openerp import fields
from openerp import models
import zmq
from zmq_rpc import ZmqRpcClient
from zmq_rpc import ZmqRpcError


log = logging.getLogger(__name__)


class XbusEmitter(models.Model):
    """Configuration profile to emit Xbus messages from Odoo.
    """

    _name = 'xbus.emitter'

    _order = 'priority ASC'

    front_url = fields.Char(
        string='Front URL',
        size=128,
        required=True,
        help='URL to the Xbus broker front.',
    )

    login = fields.Char(
        string='Login',
        size=64,
        required=True,
        help='Emitter login; must exist in the list of emitters inside Xbus.',
    )

    password = fields.Char(
        string='Password',
        size=64,
        required=True,
        help='Emitter password for the specified login.',
    )

    priority = fields.Integer(
        string='Priority',
        required=True,
        help=(
            'Priority to sort emitter configuration objects; the lowest '
            'priority is the most important.'
        ),
        default=100,
    )

    timeout = fields.Integer(
        string='Timeout',
        required=True,
        help=(
            'Time to wait during Xbus calls before throwing an error (in '
            'milliseconds).'
        ),
        default=5000,
    )

    @api.one
    def name_get(self):
        return self.id, '%s@%s' % (self.login, self.front_url)

    @api.one
    def send_items(self, event_type, items, delayed=False):
        """Send the data to Xbus, inside one event with the specified event
        type.

        :param event_type: Name of an Xbus event type.
        :type event_type: String.

        :type items: Iterable.

        :param delayed: Whether the emission should only happen after the
        database transaction is committed; a scheduled task will run the
        emission in that case. Errors, if any, will be logged.
        :type delayed: Boolean.

        :return: The result of an "end_event" Xbus API call in immediate
        (non-delayed) mode; the ID of a job in delayed mode.
        """

        if delayed:
            # Just add to the job queue.

            self.env['xbus.emitter.job'].create({
                'emitter_id': self.id,
                'event_type': event_type,
                'items': json.dumps(items),
            })

        else:
            # Not delayed; run now and return the result.

            try:
                return self._send_items_impl(event_type, items)

            except (zmq.ZMQBaseError, ZmqRpcError) as error:
                raise exceptions.Warning(_(
                    'Error when sending data to Xbus: %s'
                ) % error)

    def _send_items_impl(self, event_type, items):
        """Implementation of the Xbus data sending function that can throw.
        :param event_type: Name of an Xbus event type.
        :type event_type: String.
        :type items: Iterable.
        :raise zmq.ZMQBaseError: zmq error.
        :return: The result of an "end_event" Xbus API call.
        """

        log.debug('Establishing RPC connection...')
        conn = ZmqRpcClient(self.front_url, timeout=self.timeout)
        log.debug('RPC connection OK')
        token = conn.login(self.login, self.password)
        log.debug('Got connection token: %s', token)

        envelope_id = conn.start_envelope(token)
        log.debug('Envelope created: %s', envelope_id)

        event_id = conn.start_event(token, envelope_id, event_type, 0)
        log.debug('Event created: %s - event type: %s', event_id, event_type)

        for item in items:
            conn.send_item(
                token, envelope_id, event_id, conn.packer.pack(item)
            )
        log.debug('Data sent (event ID: %s)', event_id)

        ret = conn.end_event(token, envelope_id, event_id)
        log.debug('Event %s done: return value: %s', event_id, ret)

        conn.end_envelope(token, envelope_id)
        log.debug('Envelope %s closed', envelope_id)

        conn.logout(token)
        log.debug('Logged out; terminating')

        conn.close()
        log.debug('Done.')

        return ret
